/*
 * Copyright (c) ROSA Laboratory 2011
 * License: GPLv3
 *
 * Authors:
 *   Julia Mineeva <julia.mineeva@osinit.ru>
 *   Evgeniy Augin <evgeniy.augin@osinit.ru>
 *   Ivan Kulkov   <ivan.kulkov@osinit.ru>
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 3,
 *   or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

// local
#include "timeframedirmodel.h"

// KDE
#include <kfileitem.h>
#include <kicon.h>

// QT
#include <QIcon>
#include <QPainter>
#include <QtSvg/QSvgRenderer>

// nepomuk
#include <nepomuk/resource.h>


class DirModelNode;
class DirModelDirNode;

static KUrl cleanupUrl(const KUrl& url)
{
    KUrl u = url;
    u.cleanPath(); // remove double slashes in the path, simplify "foo/." to "foo/", etc.
    u.adjustPath(KUrl::RemoveTrailingSlash); // KDirLister does this too, so we remove the slash before comparing with the root node url.
    u.setQuery(QString());
    u.setRef(QString());
    return u;
}

// We create our own tree behind the scenes to have fast lookup from an item to its parent,
// and also to get the children of an item fast.
class DirModelNode
{
public:
    DirModelNode(DirModelDirNode* parent, const KFileItem& item ) :
        m_item(item),
        m_parent(parent),
        m_preview()
    {
    }
    // m_item is KFileItem() for the root item
    const KFileItem& item() const { return m_item; }
    void setItem(const KFileItem& item) { m_item = item; }
    DirModelDirNode* parent() const { return m_parent; }
    // linear search
    int rowNumber() const; // O(n)
    QIcon preview() const { return m_preview; }
    void setPreview( const QPixmap& pix ) {  m_preview = QIcon(); m_preview.addPixmap(pix); }
    void setPreview( const QIcon& icn ) { m_preview = icn; }

private:
    KFileItem m_item;
    DirModelDirNode* const m_parent;
    QIcon m_preview;
};

// Specialization for directory nodes
class DirModelDirNode : public DirModelNode
{
public:
    DirModelDirNode( DirModelDirNode* parent, const KFileItem& item)
        : DirModelNode( parent, item),
          m_childNodes(),
          m_childCount(KDirModel::ChildCountUnknown),
          m_populated(false)
    {}
    ~DirModelDirNode() {
        qDeleteAll(m_childNodes);
    }
    QList<DirModelNode *> m_childNodes; // owns the nodes

    // If we listed the directory, the child count is known. Otherwise it can be set via setChildCount.
    int childCount() const { return m_childNodes.isEmpty() ? m_childCount : m_childNodes.count(); }
    void setChildCount(int count) { m_childCount = count; }
    bool isPopulated() const { return m_populated; }
    void setPopulated( bool populated ) { m_populated = populated; }

    // For removing all child urls from the global hash.
    void collectAllChildUrls(KUrl::List &urls) const {
        Q_FOREACH(DirModelNode* node, m_childNodes) {
            const KFileItem& item = node->item();
            urls.append(cleanupUrl(item.url()));
            if (item.isDir())
                static_cast<DirModelDirNode*>(node)->collectAllChildUrls(urls);
        }
    }

private:
    int m_childCount:31;
    bool m_populated:1;
};

int DirModelNode::rowNumber() const
{
    if (!m_parent) return 0;
    return m_parent->m_childNodes.indexOf(const_cast<DirModelNode*>(this));
}


TimeFrameDirModel::TimeFrameDirModel(QObject *parent)
    : KDirModel(parent)
{
};

QVariant TimeFrameDirModel::data(const QModelIndex & index, int role) const
{
    if (!index.isValid())
        return QVariant();

    DirModelNode* node = static_cast<DirModelNode*>(index.internalPointer());
    const KFileItem& item(node->item());

    switch (role)
    {
    case Qt::DecorationRole:
    {
        if (index.column() == Name)
        {
            if (!node->preview().isNull())
            {
                //kDebug(7008) << item->url() << " preview found";
                return node->preview();
            }
            Q_ASSERT(!item.isNull());

            // set ours (non-standart) icon for item
            //return KIcon(item.iconName(), 0, item.overlays());
            return QIcon(":/icons/null.png");
            //return KIcon();
        }
    }
    case Qt::ToolTipRole:
    {
        Nepomuk::Resource res(item.localPath());
        int rating = static_cast<int>(res.rating());

        // tooltip with event role
        QString ttText = i18n("%1\nrating: %2", item.text(), rating);
        return ttText;
    }
    case Qt::ForegroundRole:
    {
//        return QBrush(QColor(Plasma::Theme::defaultTheme()->color(Plasma::Theme::TextColor)));
    }

    case Qt::DisplayRole:
    default:
        return KDirModel::data(index, role);
    }

    /*
    if (index.isValid()) {
        DirModelNode* node = static_cast<DirModelNode*>(index.internalPointer());
        const KFileItem& item(node->item());
        switch (role)
        {
        case Qt::DisplayRole:
            switch (index.column()) {
            case Name:
                return item.text();
            case Size:
                //
                //return KIO::convertSize(item->size());
                // Default to "file size in bytes" like in kde3's filedialog
                return KGlobal::locale()->formatNumber(item.size(), 0);
            case ModifiedTime: {
                KDateTime dt = item.time(KFileItem::ModificationTime);
                return KGlobal::locale()->formatDateTime(dt);
            }
            case Permissions:
                return item.permissionsString();
            case Owner:
                return item.user();
            case Group:
                return item.group();
            case Type:
                return item.mimeComment();
            }
            break;
        case Qt::EditRole:
            switch (index.column()) {
            case Name:
                return item.text();
            }
            break;
        case Qt::DecorationRole:
            if (index.column() == Name) {
                if (!node->preview().isNull()) {
                    //kDebug(7008) << item->url() << " preview found";
                    return node->preview();
                }
                Q_ASSERT(!item.isNull());
                //kDebug(7008) << item->url() << " overlays=" << item->overlays();
                return KIcon(item.iconName(), 0, item.overlays());
            }
            break;
        case Qt::TextAlignmentRole:
            if (index.column() == Size) {
                // use a right alignment for L2R and R2L languages
                const Qt::Alignment alignment = Qt::AlignRight | Qt::AlignVCenter;
                return int(alignment);
            }
            break;
        case Qt::ToolTipRole:
            return item.text();
        case FileItemRole:
            return QVariant::fromValue(item);
        case ChildCountRole:
            if (!item.isDir())
                return ChildCountUnknown;
            else {
                DirModelDirNode* dirNode = static_cast<DirModelDirNode *>(node);
                int count = dirNode->childCount();
                if (count == ChildCountUnknown && item.isReadable()) {
                    const QString path = item.localPath();
                    if (!path.isEmpty()) {
                        //                        slow
                        //                        QDir dir(path);
                        //                        count = dir.entryList(QDir::AllEntries|QDir::NoDotAndDotDot|QDir::System).count();
#ifdef Q_WS_WIN
                        QString s = path + QLatin1String( "\\*.*" );
                        s.replace('/', '\\');
                        count = 0;
                        WIN32_FIND_DATA findData;
                        HANDLE hFile = FindFirstFile( (LPWSTR)s.utf16(), &findData );
                        if( hFile != INVALID_HANDLE_VALUE ) {
                            do {
                                if (!( findData.cFileName[0] == '.' &&
                                       findData.cFileName[1] == '\0' ) &&
                                        !( findData.cFileName[0] == '.' &&
                                           findData.cFileName[1] == '.' &&
                                           findData.cFileName[2] == '\0' ) )
                                    ++count;
                            } while( FindNextFile( hFile, &findData ) != 0 );
                            FindClose( hFile );
                        }
#else
                        DIR* dir = ::opendir(QFile::encodeName(path));
                        if (dir) {
                            count = 0;
                            struct dirent *dirEntry = 0;
                            while ((dirEntry = ::readdir(dir))) {
                                if (dirEntry->d_name[0] == '.') {
                                    if (dirEntry->d_name[1] == '\0') // skip "."
                                        continue;
                                    if (dirEntry->d_name[1] == '.' && dirEntry->d_name[2] == '\0') // skip ".."
                                        continue;
                                }
                                ++count;
                            }
                            ::closedir(dir);
                        }
#endif
                        //kDebug(7008) << "child count for " << path << ":" << count;
                        dirNode->setChildCount(count);
                    }
                }
                return count;
            }
        case HasJobRole:
            if (d->m_jobTransfersVisible && d->m_allCurrentDestUrls.isEmpty() == false) {
                DirModelNode* node = d->nodeForIndex(index);
                const QString url = node->item().url().url();
                //return whether or not there are job dest urls visible in the view, so the delegate knows which ones to paint.
                return QVariant(d->m_allCurrentDestUrls.contains(url));
            }
        }
    }
    */
};

bool TimeFrameDirModel::setData(const QModelIndex& index, const QVariant& value, int role)
{
    if (role == Qt::DecorationRole)
    {
        KFileItem item = itemForIndex(index);
        QString mimeType = item.mimetype();

//        qDebug("KFileItem item url: %s (%s)", item.localPath().toLocal8Bit().data(), mimeType.toLocal8Bit().data());

        // set picture ">" for video event
        if (index.isValid() && (index.column() == 0) && mimeType.contains("video"))
        {
            QIcon icon = value.value<QIcon>();
            QPixmap image = icon.pixmap(256);
            QSvgRenderer svgr(QString(":/icons/pla-empty-str.svg"));

            QPainter p(&image);
            QRectF rect(image.width()/3,(image.height()-(image.width()/3))/2,image.width()/3,image.width()/3);
            svgr.render(&p,rect);
            p.end();
            icon = QIcon(image);
            QVariant newValue(icon);

            return KDirModel::setData(index, newValue, role);
        }
    }

    return KDirModel::setData(index, value, role);
};
